from typing import Iterable

import gams
import pandas as pd
from six import string_types


def all_na(x):
  """Returns bool of whether a series or scalar consists of all NAs"""
  if is_iterable(x):
    return all(pd.isna(x))
  else:
    return pd.isna(x)


def index_names_from_symbol(symbol):
  """
  Return the domain names of a GAMS symbol,
  except ['*'] cases are replaced by the name of the symbol
  and ['*',..,'*'] cases are replaced with ['index_0',..'index_n']
  """
  index_names = list(symbol.domains_as_strings)
  if index_names == ["*"]:
    return [symbol.name]
  if index_names.count("*") > 1:
    for i, name in enumerate(index_names):
      if name == "*":
        index_names[i] = f"index_{i}"
  return index_names


def index_from_symbol(symbol):
  """Return a Pandas Index based on the records and domain names of a GAMS symbol."""
  if len(symbol.domains_as_strings) > 1:
    keys = map_to_int_where_possible([rec.keys for rec in symbol])
    index = pd.MultiIndex.from_tuples(keys, names=index_names_from_symbol(symbol))
    index.name = symbol.name
  elif len(symbol.domains_as_strings) == 1:
    keys = map_to_int_where_possible([rec.keys[0] for rec in symbol])
    index = pd.Index(keys, name=index_names_from_symbol(symbol)[0])
  else:
    return None
  if isinstance(symbol, gams.GamsSet):
    index.text = symbol.text
    index.domains = symbol.domains_as_strings
    index.texts = pd.Series([rec.text for rec in symbol], index, name=symbol.name)
  return index


def symbol_is_scalar(symbol):
  return not symbol.domains_as_strings


def series_from_variable(symbol, attr="level"):
  """Get a variable symbol from the GAMS database and return an equivalent Pandas series."""
  return pd.Series([getattr(rec, attr) for rec in symbol], index_from_symbol(symbol), name=symbol.name)


def series_from_parameter(symbol):
  """Get a parameter symbol from the GAMS database and return an equivalent Pandas series."""
  return pd.Series([rec.value for rec in symbol], index_from_symbol(symbol), name=symbol.name)


def is_iterable(arg):
  return isinstance(arg, Iterable) and not isinstance(arg, string_types)


def map_lowest_level(func, x):
  """Map lowest level of zero or more nested lists."""
  if is_iterable(x):
    return [map_lowest_level(func, i) for i in x]
  else:
    return func(x)


def try_to_int(x):
  """Cast input to int if possible, else return input unchanged."""
  try:
    if str(int(x)) == str(x):
      return int(x)
    else:
      return x
  except ValueError:
    return x


def map_to_int_where_possible(iter):
  """Returns an iterable where each element is converted to an integer if possible for that element."""
  return map_lowest_level(try_to_int, iter)


def merge_symbol_records(series, symbol):
  """Convert Pandas series to records in a GAMS Symbol"""
  if isinstance(symbol, gams.GamsSet):
    attr = "text"
  elif isinstance(symbol, gams.GamsVariable):
    attr = "level"
  elif isinstance(symbol, gams.GamsParameter):
    attr = "value"
  for k, v in series.items():
    setattr(symbol.merge_record(k), attr, v)


def set_symbol_records(symbol, value):
  """Convert Pandas series to records in a GAMS Symbol"""
  if isinstance(symbol, gams.GamsSet):
    if isinstance(value, pd.Index):
      texts = getattr(value, "texts", None)
      value = texts if texts is not None else pd.Series(map(str, value), index=value)

    def add_record(symbol, k, v):
      if not all_na(v):
        symbol.add_record(k).text = str(v)
  elif isinstance(symbol, gams.GamsVariable):
    def add_record(symbol, k, v):
      if not all_na(v):
        symbol.add_record(k).level = v
  elif isinstance(symbol, gams.GamsParameter):
    def add_record(symbol, k, v):
      if not all_na(v):
        symbol.add_record(k).value = v
  else:
    TypeError(f"{type(symbol)} is not (yet) supported by gams_pandas")

  symbol.clear()
  if symbol_is_scalar(symbol):
    add_record(symbol, None, value)
  elif list(value.keys()) == [0]:  # If singleton series
    add_record(symbol, None, value[0])
  else:
    for k, v in value.items():
      add_record(symbol, map_lowest_level(str, k), v)


import plotly.graph_objects as go
import dreamtools as dt
import pandas as pd

def age_figure_3d(series,
                  start_year=None,
                  end_year=None,
                  start_age=None,
                  end_age=None,
                  title="",
                  ztitle="",
                  showscale=False,
                  **kwargs):
  if start_year is None:
    start_year = max(dt.START_YEAR, min(series.index.levels[-1]))
  if end_year is None:
    end_year = min(dt.END_YEAR, max(series.index.levels[-1]))
  if start_age is None:
    start_age = dt.START_AGE
  if end_age is None:
    end_age = dt.END_AGE

  age = list(range(start_age, end_age + 1))
  time = list(range(start_year, end_year + 1))
  value = series.loc[age, time].unstack().values
  surface = go.Surface(x=time, y=age, z=value, showscale=showscale, **kwargs)
  return go.Figure(
    surface,
    layout={
      "scene": {
        "xaxis": {"title": dt.TIME_AXIS_TITLE, "autorange": "reversed"},
        "yaxis": {"title": dt.AGE_AXIS_TITLE, "autorange": "reversed"},
        "zaxis": {"title": ztitle},
      },
      "title": {"text": title, 'x': 0.5, "y": 0.925}
    },
  )

def dummy_function(x):
  return x

def age_figure_2d(iter_series,
                  operator=None,
                  years=None,
                  start_age=None, end_age=None,
                  reference_database=None,
                  names=None,
                  function=dummy_function,
                  **kwargs
                  ):
  if isinstance(iter_series, pd.Series):
    iter_series = [iter_series]
  if years is None:
    years = list(range(dt.START_YEAR, dt.END_YEAR+1))
    if max(len(s.loc[:,years].index.unique(level=-1)) for s in iter_series) > 5:
      years = list(range(dt.START_YEAR, dt.END_YEAR + 1, 5))
  if start_age is None:
    start_age = dt.START_AGE
  if end_age is None:
    end_age = dt.END_AGE
  iter_series = [function(series.sort_index().loc[range(start_age,end_age+1), years]) for series in iter_series]
  if operator:
    if reference_database is None:
      reference_database = dt.get_reference_database()
    refs = [function(reference_database[series.name].sort_index().loc[series.index]) for series in iter_series]
    iter_series = dt.compare(iter_series, refs, operator)
  if names is not None:
    for name, series in zip(names, iter_series):
      series.name = name
  df = pd.DataFrame()
  for series in iter_series:
    series_df = series.unstack()
    if len(series_df.columns) == 1:
      df[f"{series.name}"] = series_df[series_df.columns[0]]
    else:
      for col in series_df:
        df[f"{series.name}[{col}]"] = series_df[col]
  return df.plot().update_layout(**{
    "xaxis_title_text": dt.AGE_AXIS_TITLE,
    "yaxis_title_text": dt.YAXIS_TITLE_FROM_OPERATOR.get(operator, ""),
    "legend_title": "",
    **kwargs
  })


PLOTLY = True
try:
  import plotly
except ImportError:
  PLOTLY = False

if PLOTLY:
  import pandas
  pandas.options.plotting.backend = "plotly"
  from .timeseries_analysis import *
  from .plotly_util import *
  from .age_plot import *

  import plotly.io as pio
  from .dream_plotly_template import dream_template
  pio.templates["dream"] = dream_template
  pio.templates.default = "dream"

  import dreamtools as dt
  pio.orca.config.default_scale = dt.PLOT_SCALE



from .gams_pandas import *

# Global setting controlling the default position of the time index (-1 = last index is time)
X_AXIS_INDEX = -1

# Time settings, used for plot and prt functions
START_YEAR = 1965
END_YEAR = 2100

# Age settings, currently used for plotting only
START_AGE = 0
END_AGE = 100

# Global databases
REFERENCE_DATABASE = None

# Plotly settings
LARGE_PLOT_WIDTH = 1830
PLOT_HEIGHT = 592
SMALL_PLOT_WIDTH = 897
PLOT_SCALE = 3
from .multiindex_plotly import *

# Model specific settings
from .multiindex_plotly.makro_settings import *



import plotly.offline as pyo
from plotly.subplots import make_subplots
from math import ceil


def figures_to_html(figs, filename="figures.html"):
  """Write an iter of plotly figures to an html file."""
  file = open(filename, 'w')
  file.write("<html><head></head><body>" + "\n")

  for i, fig in enumerate(figs):
    inner_html = pyo.plot(
      fig, include_plotlyjs=(i==0), output_type='div'
    )
    file.write(inner_html)

  file.write("</body></html>" + "\n")


def subplot(figures, cols=2, **kwargs):
  """Create subplot from iter of figures."""
  rows = ceil(len(figures) / cols)
  fig = make_subplots(rows=rows,
                      cols=cols,
                      subplot_titles=[f["layout"]["title"]["text"] for f in figures],
                      **kwargs
                      )
  for i, f in enumerate(figures):
    for trace in f["data"]:
      fig.add_trace(trace, row=ceil((i+1)/cols), col=1 + i % cols)
  return fig



import plotly.graph_objects as go

dream = {
    "layout": {
        "colorway": [
            "rgb(245,82,82)",
            "rgb(20,175,166)",
            "rgb(255,155,75)",
            "rgb(92,210,114)",
            "rgb(66,180,224)",
            "rgb(188,173,221)",
            "rgb(0,95,151)",
            "rgb(137,48,112)",
            "rgb(70,70,76)",
        ],
        "legend": {
            "yanchor": "bottom",
            "xanchor": "center",
            "orientation": "h",
            "y": -0.5,
            "x": 0.5,
        },
        "font": {"color": "rgb(0,0,0)", "family": "Hind", "size": 14},

        "xaxis": {
            "ticks": "outside",
            "ticklen": 10,
            "showline": True,
            "linecolor": "black",
            "automargin": False,
            "title_standoff": 0,
            "title_font_size": 14,
        },

        "yaxis": {
            "ticks": "outside",
            "ticklen": 10,
            "showline": True,
            "linecolor": "black",
            # "tickangle": -90,
            "automargin": False,
            "tickmode": "auto",
        },

        "plot_bgcolor": "rgb(230,230,232)",
        "xaxis_gridcolor": "white",
        "yaxis_gridcolor": "white",

        # "paper_bgcolor": "rgb(230,230,232)",
    },

    "data": {
        "scatter": [{
            "line": {"width": 4},
        }],
    }
}

dream["data"]["scattergl"] = dream["data"]["scatter"]

dream_template = go.layout.Template(**dream)


import os

import gams

from .gams_pandas import GamsPandasDatabase, logger


class Gdx(GamsPandasDatabase):
  """Wrapper that opens a GDX file as a GamsPandasDatabase."""

  def __init__(self, file_path, workspace=None):
    if not file_path:
      import easygui
      file_path = easygui.fileopenbox("Select reference gdx file", filetypes=["*.gdx"])
      if not file_path:
        raise ValueError("No file path was provided")
    if not os.path.splitext(file_path)[1]:
      file_path = file_path + ".gdx"
    self.abs_path = os.path.abspath(file_path)
    logger.info(f"Open GDX file from {self.abs_path}.")
    if workspace is None:
      workspace = gams.GamsWorkspace()
    database = workspace.add_database_from_gdx(self.abs_path)
    super().__init__(database)

  def export(self, path=None, relative_path=True):
    """
    Save changes to database and export database to GDX file.
    If no path is specified, the path of the linked GDX file is used, overwriting the original.
    Use relative_path=True to use export to a path relative to the directory of the originally linked GDX file.
    """
    if path is None:
      path = self.abs_path
    elif relative_path:
      path = os.path.join(os.path.dirname(self.abs_path), path)
    logger.info(f"Export GDX file to {path}.")
    super().export(path)

  def __repr__(self):
    return f"dreamtools.gams_pandas.gams_pandas.Gdx from {self.abs_path}"


import dreamtools as dt


# Åben gdx fil
gdx = dt.Gdx("test.gdx")
gdx.export("export.gdx")

# Hent sets fra gdx
s, i, t = gdx.get("s", "i", "t")

# Hent variable fra gdx
qY, qBNP, qI_s = gdx.get("qY", "qBNP", "qI_s")

# Plot variabel
qBNP.plot()

# Plot variabel i begrænset periode
qBNP.plot(2010, 2020)

# Plot variabel begrænset til et enkelt element eller set
qY["tot"].plot()
qY[s].plot()

# Plot flere variable sammen
dt.plot(qY[s], qBNP, start=2010, end=2020)

# Hvis en serie har flere dimensioner, skal vi skrive .loc for at begrænse med på mere end første dimension. Fx:
qI_s.loc[i,s]

# Vi kan stadig begænse på første dimension uden .loc
qI_s[i]

# Hvis vi begrænser første dimnension til et enkelt element kan vi derefter begrænse anden dimension uden .loc
qI_s["IB"]["bol"] == qI_s.loc["IB","bol"]
# Generelt er det en god idé at skrive .loc hvis man er i tvivl
# (det gør det eksplicit at vi vil bruge 'label based indexing' fremfor et numerisk indeks)

# Sæt globalt standard startår og slutår
dt.time(2010, 2040)

# Gang to variable sammen
vY = gdx["pY"]*qY
# Divider med variablens værdi i 2010
vY /= vY[:,'2010']
# Plot med en overskrift tilføjet
vY[s].plot(title="pY*qY, index=2010")

# Check at der eksisterer en 'images' mappe og opret den hvis den ikke gør
import os
if not os.path.exists("images"):
  os.mkdir("images")

# Plot som statisk billed, fx som svg, pdf eller png. Bemærk at mange billed-formater kan åbnes direkte i PyCharm
qBNP.plot(file="images/plot.svg")
# Dette kræver at man har Orca installeret - spørg evt. Lucas eller undertegnede for hjælp


# Skift om der renderes til browser, pdf, png etc. som standard
import plotly.io as pio
pio.renderers.default = "browser"


# Lav ny GamsPandasDatabase og alias af metoder til at lave nye symboler
db = dt.GamsPandasDatabase()
Par, Var, Set = db.create_parameter, db.create_variable, db.create_set

# Definer sets
t = Set("t", range(2000, 2020), "Årstal")
s = Set("s", ["tjenester", "fremstilling"], "Brancher")
st = Set("st", [s, t], "Branche x år dummy")
sub = Set("sub", ["tjenester"], "Subset af brancher", domains=["s"])

one2one = Set("one2one", [(2010, 2015), (2011, 2016)], "1 til 1 mapping", domains=["t", "t"])

one2many = Set("one2many",
               [("tot", "tjenester"), ("tot", "fremstilling")],
               "1 til mange mapping", domains=["*", "s"],
               )

# Definer nye parametre og variable, gerne ud fra sets
gq = Par("gq", None, "Produktivitets-vækst", 0.01)
fq = Par("fp", t, "Vækstkorrektionsfaktor", (1 + 0.01) ** (t - 2010))
d = Par("d", st, "Dummy")
y = Var("y", [s, t], "Produktion")
p = Var("p", [s, t], "Pris")

# Assignment
y["tjenester"], y["fremstilling"] = 7, 3
p = 100  # Hov, vi kom til at overskrive vores p parameter!
p = db["p"]  # Den kan heldigivs findes igen
p[:] = 100

# Beregner med symboler som deler sets sker uden problemer
y *= fq
print(p * y)


DEFAULT_SET_AGGREGATIONS = {
	"a_": ["tot",],
	"a": [10, 20, 30, 40, 50, 60, 70, 80, 90, 100],

	"c_": ['cTot',],
	"c" : ['cBil', 'cEne', 'cVar', 'cBol', 'cTje', 'cTur'],

	"x_": ['xTot',],
	"x" : ['xEne', 'xVar', 'xSoe', 'xTje', 'xTur'],

	"g_": ['gTot',],
	"g" : ['g',],

	"s_": ['tot',],
	"sp": ['tje', 'fre', 'byg', 'soe', 'bol', 'lan', 'ene', 'udv'],
	"s" : ['tje', 'fre', 'byg', 'soe', 'bol', 'lan', 'ene', 'udv', 'off'],
	"r_": ['rTot',],
	"r" : ['tje', 'fre', 'byg', 'soe', 'bol', 'lan', 'ene', 'udv', 'off'],

	"i_": ['iTot',],
	"i" : ['iM', 'iB', 'iL'],

	"k_": ['iM', 'iB'],
	"k" : ['iM', 'iB'],

  "portf": ['Net',],
  "portf_": ['Net',],

  "scen_": ['GovSpending', 'IntRate', 'ForeignDemand'],
}

AGE_AXIS_TITLE = "Alder"
TIME_AXIS_TITLE = "År"

YAXIS_TITLE_FROM_OPERATOR = {
	"pq": "Pct.-ændringer relativt til grundforløb",
	"q": "Ændring relativt til grundforløb",
	"m": "Forskel fra grundfoløb",
}
